# Local imports:
from . import EnhCanvas, EnhFrame
from ..utils import Orientation

# Standard libraries:
from typing import Optional, Tuple, List, Any
from tkinter import Event, Misc


class EnhScrollableFrame(EnhFrame):
    in_frame_id: Optional[int] = None
    in_frame: Optional[EnhFrame] = None
    orientation: Orientation = None

    def __init__(
            self,
            master: Optional[Misc] = None,
            *,
            background: Optional[str] = "transparent",
            orient: Orientation = "vertical",
            **standard_options: Any) -> None:
        super().__init__(master, background=background, **standard_options)
        self.create_widgets()
        self.configure(orient=orient)

    def create_widgets(self) -> None:
        self.canvas: EnhCanvas = EnhCanvas(self, height=0, width=0)
        self.canvas.bind(sequence="<MouseWheel>", func=self.on_mouse_wheel, add=True)
        self.canvas.bind(sequence="<Configure>", func=self.on_configure, add=True)
        self.canvas.bind(sequence="<Map>", func=self.on_map, add=True)
        self.in_frame: EnhFrame = EnhFrame(self.canvas)
        self.in_frame.bind(sequence="<MouseWheel>", func=self.on_mouse_wheel, add=True)
        self.in_frame.bind(sequence="<Configure>", func=self.on_configure, add=True)
        coords: Tuple[int, int] = 0, 0
        self.in_frame_id = self.canvas.create_window(coords, anchor="nw", window=self.in_frame)

    def on_mouse_wheel(self, event: Event) -> None:
        scroll_amount: int = event.delta // 120
        self.canvas.yview_scroll(number=-scroll_amount, what="units") if self.orientation == "vertical" \
            else self.canvas.xview_scroll(number=-scroll_amount, what="units")

    def on_configure(self, event: Event) -> None:
        scroll_coordinates: Tuple[int, int, int, int] = self.canvas.bbox("all")
        scroll_coordinates: List[int] = list(scroll_coordinates)
        canvas_end_x, canvas_end_y = scroll_coordinates[2:]
        canvas_width, canvas_height = event.width, event.height
        scroll_coordinates[2] = max(canvas_end_x, canvas_width)
        scroll_coordinates[3] = max(canvas_end_y, canvas_height)
        self.canvas.configure(scrollregion=scroll_coordinates)
        is_vertical: bool = self.orientation == "vertical"
        self.canvas.itemconfigure(
            tagOrId=self.in_frame_id,
            height=0 if is_vertical else canvas_height,
            width=canvas_width if is_vertical else 0)

    def on_map(self, event: Optional[Event] = None) -> None:
        for child_widget in self.winfo_children():
            if child_widget == self.canvas:
                continue
            self.bind_mouse_wheel(widget=child_widget)

    def bind_mouse_wheel(self, widget: Misc) -> None:
        widget.bind(sequence="<MouseWheel>", func=self.on_mouse_wheel, add=True)
        for child_widget in widget.winfo_children():
            self.bind_mouse_wheel(widget=child_widget)

    def configure(self, **standard_options: Any) -> None:
        orientation: Orientation = standard_options.pop("orient", None)
        if orientation is not None:
            self.orientation = orientation or "vertical"
            is_vertical: bool = self.orientation == "vertical"
            self.canvas.pack_configure(expand=True, fill="both", side="left" if is_vertical else "top")
        super().configure(**standard_options)
    config = configure



# standard libraries:
from typing import Optional, Callable, Final, Dict, List
from tkinter import Event, Misc, Wm, TclError
from abc import ABC, abstractmethod
from ctypes import WinDLL, windll

# window resize cursor names:
WINDOW_CURSOR_NAMES: Final[Dict[int, str]] = {
    0: "size_nw_se",
    1: "size_ne_sw",
    2: "sb_v_double_arrow",
    3: "size_ne_sw",
    4: "size_nw_se",
    5: "sb_v_double_arrow",
    6: "sb_h_double_arrow",
    7: "sb_h_double_arrow"}


class WindowInterface(ABC):
    @abstractmethod
    def retrieve_window_handle(self) -> int: ...

    @abstractmethod
    def hide_titlebar(self) -> None: ...

    @abstractmethod
    def show_titlebar(self) -> None: ...

    @abstractmethod
    def center_window(self, *, width: Optional[int] = None, height: Optional[int] = None) -> None: ...

    @abstractmethod
    def minimize_window(self) -> None: ...
    minimize = minimize_window


class WindowIndexes:
    WINDOW_STYLE: Final[int] = -16
    EXTENDED_WINDOW_STYLE: Final[int] = -20


class WindowStyles:
    MINIMIZE_BOX: Final[int] = 0x00020000
    FORCE_ONTO_TASKBAR: Final[int] = 0x00040000


class WindowStates:
    NORMAL: Final[str] = 1
    ZOOMED: Final[str] = 3
    SHOWN: Final[int] = 5
    MINIMIZED: Final[int] = 6


class WindowResizeUtils:
    _events: List[Dict[str, str]] = []
    _window: Optional["WindowUtils"] = None

    last_cursor_index: Optional[int] = None
    started_x: Optional[int] = None
    started_y: Optional[int] = None

    @staticmethod
    def get_cursor_index(event: Event, window: "WindowUtils") -> int:
        is_window: bool = event.widget == window
        if not is_window:
            return -1
        is_maximized: bool = window.wm_state() == WindowStates.ZOOMED
        if is_maximized:
            return -1
        is_top: bool = event.y <= 5
        is_left: bool = event.x <= 5
        is_right: bool = event.x >= window.winfo_width() - 5
        is_bottom: bool = event.y >= window.winfo_height() - 5
        is_width_resizable, is_height_resizable = window.wm_resizable()
        if is_height_resizable and is_top:
            return 0 if is_left else 1 if is_right else 2
        if is_height_resizable and is_bottom:
            return 3 if is_left else 4 if is_right else 5
        return -1 if not is_width_resizable else 6 if is_left else 7 if is_right else -1

    def __init__(self, master: Optional["WindowUtils"] = None) -> None:
        assert isinstance(master, WindowUtils), "The parent master must be a WindowUtils instance."
        self._window = master
        self.attach_events()

    def attach_events(self) -> None:
        self.attach_event(sequence="<Motion>", function=self.on_motion)
        self.attach_event(sequence="<Button-1>", function=self.start_resize)
        self.attach_event(sequence="<B1-Motion>", function=self.on_resize)
        self.attach_event(sequence="<ButtonRelease-1>", function=self.stop_resize)

    def attach_event(self, sequence: str, function: Callable[[Event], object]) -> None:
        function_id: str = self._window.bind(sequence, func=function, add=True)
        new_event: Dict[str, str] = {"sequence": sequence, "function_id": function_id}
        self._events.append(new_event)

    def on_motion(self, event: Event) -> None:
        cursor_index: int = self.get_cursor_index(event, window=self._window)
        if self.last_cursor_index == cursor_index:
            return None
        self.last_cursor_index = cursor_index
        cursor_name: str = WINDOW_CURSOR_NAMES.get(cursor_index, "")
        self._window.configure(cursor=cursor_name)

    def start_resize(self, event: Event) -> None:
        is_default_cursor: bool = self.last_cursor_index == -1
        if is_default_cursor:
            return None
        self.started_x, self.started_y = event.x, event.y

    def on_resize(self, event: Event) -> None:
        if self.started_x is None or self.started_y is None:
            return None
        new_x: int = self._window.winfo_x()
        new_y: int = self._window.winfo_y()
        new_width: int = self._window.winfo_width()
        new_height: int = self._window.winfo_height()
        minimum_width, minimum_height = self._window.wm_minsize()
        is_width_resizable, is_height_resizable = self._window.wm_resizable()
        if is_height_resizable:
            mouse_y: int = event.y
            # top movement.
            if self.last_cursor_index in (0, 1, 2):
                minimum_y: int = new_y + new_height - minimum_height
                new_y += mouse_y - self.started_y
                new_y = min(minimum_y, new_y)
                new_height -= mouse_y - self.started_y
            # bottom movement.
            if self.last_cursor_index in (3, 4, 5):
                if mouse_y >= minimum_height:
                    variable_name: int = mouse_y - self.started_y
                    self.started_y = mouse_y
                    new_height += variable_name
                else:
                    new_height = minimum_height
        if is_width_resizable:
            mouse_x: int = event.x
            # left movement.
            if self.last_cursor_index in (0, 3, 6):
                minimum_x: int = new_x + new_width - minimum_width
                new_x += mouse_x - self.started_x
                new_x = min(minimum_x, new_x)
                new_width -= mouse_x - self.started_x
            # right movement.
            if self.last_cursor_index in (1, 4, 7):
                if mouse_x >= minimum_width:
                    variable_name: int = mouse_x - self.started_x
                    self.started_x = mouse_x
                    new_width += variable_name
                else:
                    new_width = minimum_width
        new_width = max(minimum_width, new_width)
        new_height = max(minimum_height, new_height)
        new_geometry: str = "{}x{}+{}+{}".format(new_width, new_height, new_x, new_y)
        self._window.wm_geometry(newGeometry=new_geometry)

    def stop_resize(self, event: Event) -> None:
        self.started_x = self.started_y = None
        self.on_motion(event)

    def __del__(self) -> None:
        try:
            if not self._window.winfo_exists():
                return None
            for event in self._events:
                sequence, function_id = event.values()
                self._window.unbind(sequence, funcid=function_id)
            self._window.configure(cursor="")
        except TclError:
            return None


class WindowUtils(WindowInterface, Misc, Wm):
    _resize_utils: Optional[WindowResizeUtils] = None
    _user32_api: WinDLL = windll.user32

    def retrieve_window_handle(self) -> int:
        self.update_idletasks()
        window_id: int = self.winfo_id()
        window_handle: int = self._user32_api.GetParent(window_id)
        return window_handle

    def hide_titlebar(self) -> None:
        self.wm_overrideredirect(boolean=True)
        self._resize_utils = WindowResizeUtils(self)
        window_handle: int = self.retrieve_window_handle()
        self._user32_api.SetWindowLongPtrW(
            window_handle,
            WindowIndexes.WINDOW_STYLE,
            WindowStyles.MINIMIZE_BOX)
        self._user32_api.SetWindowLongPtrW(
            window_handle,
            WindowIndexes.EXTENDED_WINDOW_STYLE,
            WindowStyles.FORCE_ONTO_TASKBAR)
        is_window_hidden: bool = self.wm_state() == "withdrawn"
        if is_window_hidden:
            return None
        self._user32_api.ShowWindow(window_handle, WindowStates.SHOWN)

    def show_titlebar(self) -> None:
        if self._resize_utils is not None:
            self._resize_utils.__del__()
            self._resize_utils = None
        self.wm_overrideredirect(boolean=False)

    def center_window(self, *, width: Optional[int] = None, height: Optional[int] = None) -> None:
        if not width or not height:
            self.update_idletasks()
        screen_width: int = self.winfo_screenwidth()
        screen_height: int = self.winfo_screenheight()
        window_width: int = width or self.winfo_width()
        window_height: int = height or self.winfo_height()
        center_x: int = (screen_width - window_width) // 2
        center_y: int = (screen_height - window_height) // 2
        new_geometry: str = "{}x{}+{}+{}".format(window_width, window_height, center_x, center_y)
        self.wm_geometry(newGeometry=new_geometry)

    def minimize_window(self) -> None:
        is_titlebar_hidden: bool = self.wm_overrideredirect()
        if not is_titlebar_hidden:
            return self.wm_iconify()
        window_handle: int = self.retrieve_window_handle()
        self._user32_api.ShowWindow(window_handle, WindowStates.MINIMIZED)
    minimize = minimize_window



# local imports:
from ..components import TransparentMisc

# standard libraries:
from tkinter import Listbox, Canvas, Frame, Text


class EnhListbox(TransparentMisc, Listbox):
    pass


class EnhCanvas(TransparentMisc, Canvas):
    pass


class EnhFrame(TransparentMisc, Frame):
    pass


class EnhText(TransparentMisc, Text):
    pass



# local imports:
from ..utils import PathDescription, Size
from ..components import TransparentMisc

# standard libraries:
from typing import Optional, Tuple, Any
from tkinter import Event, Label, Misc
from threading import Thread
from logging import warning
from os.path import exists

# third-party libraries:
from PIL.Image import Image, open as open_image
from PIL.ImageTk import PhotoImage


class EnhImage(TransparentMisc, Label):
    _image: Optional[Image] = None
    _image_size: Optional[Tuple[int, int]] = None
    _photo_image: Optional[PhotoImage] = None
    _scheduled_resize: Optional[str] = None

    @staticmethod
    def get_image_size(container: Misc, image: Image) -> Tuple[int, int]:
        container_width: int = container.winfo_width()
        container_height: int = container.winfo_height()
        image_width, image_height = image.size
        scaling_factor: int = min(container_width / image_width, container_height / image_height)
        new_image_width: int = int(image_width * scaling_factor)
        new_image_height: int = int(image_height * scaling_factor) - 4
        return 1 if new_image_width <= 0 else new_image_width, 1 if new_image_height <= 0 else new_image_height

    def __init__(self, master: Optional[Misc] = None, *, imagepath: PathDescription = None, imagesize: Size = None, **standard_options: Any) -> None:
        standard_options.pop("text", None)
        super().__init__(master, **standard_options)
        self.configure(imagepath=imagepath, imagesize=imagesize)
        self.bind(sequence="<Configure>", func=self.schedule_resize, add=True)

    def schedule_resize(self, event: Optional[Event] = None) -> None:
        if self._image is None:
            return None
        if self._scheduled_resize is not None:
            self.after_cancel(id=self._scheduled_resize)
            self._scheduled_resize = None
        thread: Thread = Thread(target=self.auto_resize, daemon=True)
        self._scheduled_resize = self.after(ms=20, func=thread.start)

    def auto_resize(self) -> None:
        if self._image is None:
            return None
        new_image: Image = self._image.copy()
        new_size: Tuple[int, int] = self._image_size or self.get_image_size(self, image=new_image)
        if new_size == new_image.size:
            self._scheduled_resize = None
            return None
        resized_image: Image = new_image.resize(size=new_size)
        photo_image: PhotoImage = PhotoImage(image=resized_image)
        Misc.configure(self, image=photo_image)
        self._photo_image = photo_image
        self._scheduled_resize = None

    def configure(self, **standard_options: Any) -> Any:
        standard_options.pop("text", None)
        image_path: Optional[str] = standard_options.pop("imagepath", None)
        if image_path is not None and image_path and exists(path=image_path):
            self._image = open_image(fp=image_path)
            self.schedule_resize(event=None)
        elif image_path is not None and not exists(path=image_path):
            warning(msg="Image path not found.")
            self._image = None
            self._photo_image = None
        image_size: Size = standard_options.pop("imagesize", None)
        if self._image_size != image_size:
            self._image_size = image_size
            self.schedule_resize(event=None)
        return super().configure(**standard_options)
    config = configure



# standard libraries:
from typing import TypeAlias, Optional, Literal, Union, Tuple, List, Any
from tkinter.font import Font
from _tkinter import Tcl_Obj
from pathlib import Path

# path aliases:
PathDescription: TypeAlias = Optional[Union[Path, str]]

# other type aliases:
FontDescription: TypeAlias = (
    str
    | Font
    | List[Any]
    | Tuple[str]
    | Tuple[str, int]
    | Tuple[str, int, str]
    | Tuple[str, int, List[str] | Tuple[str, ...]]
    | Tcl_Obj)
Orientation: TypeAlias = Optional[Literal["vertical", "horizontal"]]
Size: TypeAlias = Optional[Tuple[int, int]]



# standard libraries:
from typing import Generator, Optional, Union, Tuple
from struct import unpack
from logging import error

# third-party libraries:
from numpy import ndarray, linspace, int64


class ColorUtils:
    @staticmethod
    def clamp(value: Union[int64, int]) -> int:
        value_int: int = int(value)
        clamped_value: int = min(255, value_int)
        return max(0, clamped_value)

    @staticmethod
    def is_character_hex(character: str) -> bool:
        return character.isdigit() or ("a" <= character.lower() <= "f")

    @staticmethod
    def is_hex_code_valid(hex_code: str) -> bool:
        if not hex_code.startswith("#"):
            return False
        hex_characters: str = hex_code[1:]
        return all(ColorUtils.is_character_hex(character) for character in hex_characters)

    @staticmethod
    def convert_hex_to_rgb(hex_code: str) -> Optional[Tuple[int, int, int]]:
        if not ColorUtils.is_hex_code_valid(hex_code):
            error_message: str = "Invalid hex-code: \"{}\".".format(hex_code)
            return error(msg=error_message)
        hex_code_stripped: str = hex_code.lstrip("#")
        byte_data: bytes = bytes.fromhex(hex_code_stripped)
        rgb_values: Tuple[int, int, int] = unpack("BBB", byte_data)
        return rgb_values

    @staticmethod
    def convert_rgb_to_hex(red: int, green: int, blue: int) -> str:
        hex_code: str = "#{:02x}{:02x}{:02x}".format(red, green, blue)
        return hex_code[:7]

    @staticmethod
    def adjust_hex_code(hex_code: str, *, brightness_factor: Optional[float] = 1.) -> str:
        if not ColorUtils.is_hex_code_valid(hex_code):
            error_message: str = "Invalid hex-code: \"{}\".".format(hex_code)
            return error(msg=error_message)
        red, green, blue = ColorUtils.convert_hex_to_rgb(hex_code)
        new_rgb_color: Tuple[int, int, int] = (
            red * brightness_factor,
            green * brightness_factor,
            blue * brightness_factor)
        new_red, new_green, new_blue = map(ColorUtils.clamp, new_rgb_color)
        return ColorUtils.convert_rgb_to_hex(red=new_red, green=new_green, blue=new_blue)

    @staticmethod
    def generate_palette_colors(start_hex_code: str, end_hex_code: str, *, color_amount: Optional[int] = 2) -> Generator[str, None, str]:
        if not ColorUtils.is_hex_code_valid(hex_code=start_hex_code):
            error_message: str = "Invalid start hex-code: \"{}\".".format(start_hex_code)
            return error(msg=error_message)
        if not ColorUtils.is_hex_code_valid(hex_code=end_hex_code):
            error_message: str = "Invalid end hex-code: \"{}\".".format(end_hex_code)
            return error(msg=error_message)
        if color_amount < 2:
            return error(msg="The color amount must be at least 2.")
        start_rgb_values: Tuple[int, int, int] = ColorUtils.convert_hex_to_rgb(start_hex_code)
        end_rgb_values: Tuple[int, int, int] = ColorUtils.convert_hex_to_rgb(end_hex_code)
        interpolated_red: ndarray = linspace(
            start=start_rgb_values[0],
            stop=end_rgb_values[0],
            num=color_amount,
            dtype=int)
        interpolated_green: ndarray = linspace(
            start=start_rgb_values[1],
            stop=end_rgb_values[1],
            num=color_amount,
            dtype=int)
        interpolated_blue: ndarray = linspace(
            start=start_rgb_values[2],
            stop=end_rgb_values[2],
            num=color_amount,
            dtype=int)
        for integer in range(color_amount):
            rgb_values: Tuple[int64, ...] = interpolated_red[integer], interpolated_green[integer], interpolated_blue[integer]
            hex_code: str = ColorUtils.convert_rgb_to_hex(*rgb_values)
            yield hex_code

    @staticmethod
    def is_hex_code_darker(hex_code: str) -> bool:
        if not ColorUtils.is_hex_code_valid(hex_code):
            error_message: str = "Invalid hex-code: \"{}\".".format(hex_code)
            return error(msg=error_message)
        red, green, blue = ColorUtils.convert_hex_to_rgb(hex_code)
        luminance: float = (red * .299 + green * .587 + blue * .114) / 255
        return luminance <= .52
